/*
 * Copyright 2012-2019 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package SpringCore

import (
	"errors"
	"fmt"
	"io"
	"reflect"
	"sort"
	"strings"
	"time"

	"github.com/go-spring/go-spring-parent/spring-logger"
	"github.com/go-spring/go-spring-parent/spring-utils"
	"github.com/spf13/cast"
	"github.com/spf13/viper"
)

func init() {

	// 注册时长转换函数 string -> time.Duration converter
	// time units are "ns", "us" (or "µs"), "ms", "s", "m", "h"。
	RegisterTypeConverter(func(s string) time.Duration {
		r, err := cast.ToDurationE(s)
		SpringUtils.Panic(err).When(err != nil)
		return r
	})

	// 注册日期转换函数 string -> time.Time converter
	// 支持非常多的日期格式，参见 cast.StringToDate。
	RegisterTypeConverter(func(s string) time.Time {
		r, err := cast.ToTimeE(s)
		SpringUtils.Panic(err).When(err != nil)
		return r
	})
}

// defaultProperties Properties 的默认实现
type defaultProperties struct {
	properties map[string]interface{}
}

// NewDefaultProperties defaultProperties 的构造函数
func NewDefaultProperties() *defaultProperties {
	return &defaultProperties{
		properties: make(map[string]interface{}),
	}
}

func (p *defaultProperties) readProperties(reader func(*viper.Viper) error) {

	v := viper.New()
	err := reader(v)
	SpringUtils.Panic(err).When(err != nil)

	keys := v.AllKeys()
	sort.Strings(keys)

	for _, key := range keys {
		val := v.Get(key)
		p.SetProperty(key, val)
		SpringLogger.Tracef("%s=%v", key, val)
	}
}

// LoadProperties 加载属性配置文件，支持 properties、yaml 和 toml 三种文件格式。
func (p *defaultProperties) LoadProperties(filename string) {
	SpringLogger.Debug("load properties from file: ", filename)

	p.readProperties(func(v *viper.Viper) error {
		v.SetConfigFile(filename)
		return v.ReadInConfig()
	})
}

// ReadProperties 读取属性配置文件，支持 properties、yaml 和 toml 三种文件格式。
func (p *defaultProperties) ReadProperties(reader io.Reader, configType string) {
	SpringLogger.Debug("load properties from reader type: ", configType)

	p.readProperties(func(v *viper.Viper) error {
		v.SetConfigType(configType)
		return v.ReadConfig(reader)
	})
}

// GetProperty 返回 keys 中第一个存在的属性值，属性名称统一转成小写。
func (p *defaultProperties) GetProperty(keys ...string) interface{} {
	for _, key := range keys {
		if v, ok := p.properties[strings.ToLower(key)]; ok {
			return v
		}
	}
	return nil
}

// GetBoolProperty 返回 keys 中第一个存在的布尔型属性值，属性名称统一转成小写。
func (p *defaultProperties) GetBoolProperty(keys ...string) bool {
	return cast.ToBool(p.GetProperty(keys...))
}

// GetIntProperty 返回 keys 中第一个存在的有符号整型属性值，属性名称统一转成小写。
func (p *defaultProperties) GetIntProperty(keys ...string) int64 {
	return cast.ToInt64(p.GetProperty(keys...))
}

// GetUintProperty 返回 keys 中第一个存在的无符号整型属性值，属性名称统一转成小写。
func (p *defaultProperties) GetUintProperty(keys ...string) uint64 {
	return cast.ToUint64(p.GetProperty(keys...))
}

// GetFloatProperty 返回 keys 中第一个存在的浮点型属性值，属性名称统一转成小写。
func (p *defaultProperties) GetFloatProperty(keys ...string) float64 {
	return cast.ToFloat64(p.GetProperty(keys...))
}

// GetStringProperty 返回 keys 中第一个存在的字符串型属性值，属性名称统一转成小写。
func (p *defaultProperties) GetStringProperty(keys ...string) string {
	return cast.ToString(p.GetProperty(keys...))
}

// GetDurationProperty 返回 keys 中第一个存在的 Duration 类型属性值，属性名称统一转成小写。
func (p *defaultProperties) GetDurationProperty(keys ...string) time.Duration {
	return cast.ToDuration(p.GetProperty(keys...))
}

// GetTimeProperty 返回 keys 中第一个存在的 Time 类型的属性值，属性名称统一转成小写。
func (p *defaultProperties) GetTimeProperty(keys ...string) time.Time {
	return cast.ToTime(p.GetProperty(keys...))
}

// SetProperty 设置属性值，属性名称统一转成小写。
func (p *defaultProperties) SetProperty(key string, value interface{}) {
	p.properties[strings.ToLower(key)] = value
}

// GetDefaultProperty 返回属性值，如果没有找到则使用指定的默认值，属性名称统一转成小写。
func (p *defaultProperties) GetDefaultProperty(key string, def interface{}) (interface{}, bool) {
	if v, ok := p.properties[strings.ToLower(key)]; ok {
		return v, true
	}
	return def, false
}

// GetPrefixProperties 返回指定前缀的属性值集合，属性名称统一转成小写。
func (p *defaultProperties) GetPrefixProperties(prefix string) map[string]interface{} {
	prefix = strings.ToLower(prefix)
	result := make(map[string]interface{})
	for k, v := range p.properties {
		if k == prefix || strings.HasPrefix(k, prefix+".") {
			result[k] = v
		}
	}
	return result
}

// GetProperties 返回所有的属性值，属性名称统一转成小写。
func (p *defaultProperties) GetProperties() map[string]interface{} {
	return p.properties
}

// bindOption 属性值绑定可选项
type bindOption struct {
	propNamePrefix string // 属性名前缀
	fullPropName   string // 完整属性名
	fieldName      string // 结构体字段的名称
	allAccess      bool   // 私有字段是否绑定
}

// bindStruct 对结构体进行属性值绑定
func bindStruct(p Properties, v reflect.Value, opt bindOption) {
	t := v.Type()
	for i := 0; i < t.NumField(); i++ {
		ft := t.Field(i)
		fv := v.Field(i)

		// 可能会开放私有字段
		fv = SpringUtils.ValuePatchIf(fv, opt.allAccess)
		subFieldName := opt.fieldName + ".$" + ft.Name

		// 字段的绑定可选项
		subOpt := bindOption{
			propNamePrefix: opt.propNamePrefix,
			fullPropName:   opt.fullPropName,
			fieldName:      subFieldName,
			allAccess:      opt.allAccess,
		}

		if tag, ok := ft.Tag.Lookup("value"); ok {
			bindStructField(p, fv, tag, subOpt)
			continue
		}

		// 匿名嵌套需要处理，不是结构体的具名字段无需处理
		if ft.Anonymous || ft.Type.Kind() == reflect.Struct {
			bindStruct(p, fv, subOpt)
		}
	}
}

// bindStructField 对结构体的字段进行属性绑定
func bindStructField(p Properties, v reflect.Value, str string, opt bindOption) {

	// 检查 tag 语法是否正确
	if !(strings.HasPrefix(str, "${") && strings.HasSuffix(str, "}")) {
		panic(fmt.Errorf("%s 属性绑定的语法发生错误", opt.fieldName))
	}

	// 指针不能作为属性绑定的目标
	if v.Kind() == reflect.Ptr {
		panic(fmt.Errorf("%s 属性绑定的目标不能是指针", opt.fieldName))
	}

	ss := strings.Split(str[2:len(str)-1], ":=")

	var (
		key string
		def interface{}
	)

	key = ss[0]

	// 此处使用最短属性名
	if opt.fullPropName == "" {
		opt.fullPropName = key
	} else if key != "" {
		opt.fullPropName = opt.fullPropName + "." + key
	}

	// 属性名如果有前缀要加上前缀
	if opt.propNamePrefix != "" {
		key = opt.propNamePrefix + "." + key
	}

	if len(ss) > 1 {
		def = ss[1]
	}

	bindValue(p, v, key, def, opt)
}

func getPropertyValue(p Properties, k reflect.Kind, key string, def interface{}, opt bindOption) interface{} {

	// 首先获取精确匹配的属性值
	if val, ok := p.GetDefaultProperty(key, nil); ok {
		return val
	}

	// Map 和 Struct 类型获取具有相同前缀的属性值
	if k == reflect.Map || k == reflect.Struct {
		if prefixValue := p.GetPrefixProperties(key); len(prefixValue) > 0 {
			return prefixValue
		}
	}

	// 最后使用默认值
	if def != nil {
		return def
	}

	panic(fmt.Errorf("%s properties \"%s\" not config", opt.fieldName, opt.fullPropName))
}

// bindValue 对任意 value 进行属性绑定
func bindValue(p Properties, v reflect.Value, key string, def interface{}, opt bindOption) {

	t := v.Type()
	k := t.Kind()

	// 存在值类型转换器的情况下结构体优先使用属性值绑定
	if fn, ok := typeConverters[t]; ok {
		propValue := getPropertyValue(p, k, key, def, opt)
		fnValue := reflect.ValueOf(fn)
		out := fnValue.Call([]reflect.Value{reflect.ValueOf(propValue)})
		v.Set(out[0])
		return
	}

	if k == reflect.Struct {
		if def == nil {
			bindStruct(p, v, bindOption{
				propNamePrefix: key,
				fullPropName:   opt.fullPropName,
				fieldName:      opt.fieldName,
				allAccess:      opt.allAccess,
			})
			return
		} else { // 前面已经校验过是否存在值类型转换器
			panic(fmt.Errorf("%s 结构体字段不能指定默认值", opt.fieldName))
		}
	}

	propValue := getPropertyValue(p, k, key, def, opt)

	switch k {
	case reflect.Uint64, reflect.Uint32, reflect.Uint16, reflect.Uint8, reflect.Uint:
		if u, err := cast.ToUint64E(propValue); err == nil {
			v.SetUint(u)
		} else {
			panic(fmt.Errorf("property value %s isn't uint type", opt.fullPropName))
		}
	case reflect.Int64, reflect.Int32, reflect.Int16, reflect.Int8, reflect.Int:
		if i, err := cast.ToInt64E(propValue); err == nil {
			v.SetInt(i)
		} else {
			panic(fmt.Errorf("property value %s isn't int type", opt.fullPropName))
		}
	case reflect.Float64, reflect.Float32:
		if f, err := cast.ToFloat64E(propValue); err == nil {
			v.SetFloat(f)
		} else {
			panic(fmt.Errorf("property value %s isn't float type", opt.fullPropName))
		}
	case reflect.String:
		if s, err := cast.ToStringE(propValue); err == nil {
			v.SetString(s)
		} else {
			panic(fmt.Errorf("property value %s isn't string type", opt.fullPropName))
		}
	case reflect.Bool:
		if b, err := cast.ToBoolE(propValue); err == nil {
			v.SetBool(b)
		} else {
			panic(fmt.Errorf("property value %s isn't bool type", opt.fullPropName))
		}
	case reflect.Slice:
		elemType := v.Type().Elem()
		elemKind := elemType.Kind()

		// 如果是字符串的话，尝试按照逗号进行切割
		if s, ok := propValue.(string); ok {
			propValue = strings.Split(s, ",")
		}

		// 处理使用类型转换器的场景
		if fn, ok := typeConverters[elemType]; ok {
			if s0, err := cast.ToStringSliceE(propValue); err == nil {
				sv := reflect.MakeSlice(t, len(s0), len(s0))
				fnValue := reflect.ValueOf(fn)
				for i, iv := range s0 {
					res := fnValue.Call([]reflect.Value{reflect.ValueOf(iv)})
					sv.Index(i).Set(res[0])
				}
				v.Set(sv)
				return
			} else {
				panic(fmt.Errorf("property value %s isn't []string type", opt.fullPropName))
			}
		}

		switch elemKind {
		case reflect.Uint64:
			if i, err := SpringUtils.ToUint64SliceE(propValue); err == nil {
				v.Set(reflect.ValueOf(i))
			} else {
				panic(fmt.Errorf("property value %s isn't []uint64 type", opt.fullPropName))
			}
		case reflect.Uint32:
			if i, err := SpringUtils.ToUint32SliceE(propValue); err == nil {
				v.Set(reflect.ValueOf(i))
			} else {
				panic(fmt.Errorf("property value %s isn't []uint32 type", opt.fullPropName))
			}
		case reflect.Uint16:
			if i, err := SpringUtils.ToUint16SliceE(propValue); err == nil {
				v.Set(reflect.ValueOf(i))
			} else {
				panic(fmt.Errorf("property value %s isn't []uint16 type", opt.fullPropName))
			}
		case reflect.Uint8:
			if i, err := SpringUtils.ToUint8SliceE(propValue); err == nil {
				v.Set(reflect.ValueOf(i))
			} else {
				panic(fmt.Errorf("property value %s isn't []uint8 type", opt.fullPropName))
			}
		case reflect.Uint:
			if i, err := SpringUtils.ToUintSliceE(propValue); err == nil {
				v.Set(reflect.ValueOf(i))
			} else {
				panic(fmt.Errorf("property value %s isn't []uint type", opt.fullPropName))
			}
		case reflect.Int64:
			if i, err := SpringUtils.ToInt64SliceE(propValue); err == nil {
				v.Set(reflect.ValueOf(i))
			} else {
				panic(fmt.Errorf("property value %s isn't []int64 type", opt.fullPropName))
			}
		case reflect.Int32:
			if i, err := SpringUtils.ToInt32SliceE(propValue); err == nil {
				v.Set(reflect.ValueOf(i))
			} else {
				panic(fmt.Errorf("property value %s isn't []int32 type", opt.fullPropName))
			}
		case reflect.Int16:
			if i, err := SpringUtils.ToInt16SliceE(propValue); err == nil {
				v.Set(reflect.ValueOf(i))
			} else {
				panic(fmt.Errorf("property value %s isn't []int16 type", opt.fullPropName))
			}
		case reflect.Int8:
			if i, err := SpringUtils.ToInt8SliceE(propValue); err == nil {
				v.Set(reflect.ValueOf(i))
			} else {
				panic(fmt.Errorf("property value %s isn't []int8 type", opt.fullPropName))
			}
		case reflect.Int:
			if i, err := SpringUtils.ToIntSliceE(propValue); err == nil {
				v.Set(reflect.ValueOf(i))
			} else {
				panic(fmt.Errorf("property value %s isn't []int type", opt.fullPropName))
			}
		case reflect.Float64, reflect.Float32:
			panic(errors.New("暂未支持"))
		case reflect.String:
			if i, err := cast.ToStringSliceE(propValue); err == nil {
				v.Set(reflect.ValueOf(i))
			} else {
				panic(fmt.Errorf("property value %s isn't []string type", opt.fullPropName))
			}
		case reflect.Bool:
			if b, err := cast.ToBoolSliceE(propValue); err == nil {
				v.Set(reflect.ValueOf(b))
			} else {
				panic(fmt.Errorf("property value %s isn't []bool type", opt.fullPropName))
			}
		default:
			// 处理结构体字段的场景
			if s, ok := propValue.([]interface{}); ok {
				result := reflect.MakeSlice(t, len(s), len(s))
				for i, si := range s {
					if sv, err := cast.ToStringMapE(si); err == nil {
						ev := reflect.New(elemType)
						subFullPropName := fmt.Sprintf("%s[%d]", key, i)
						bindStruct(&defaultProperties{sv}, ev.Elem(), bindOption{
							fullPropName: subFullPropName,
							fieldName:    opt.fieldName,
							allAccess:    opt.allAccess,
						})
						result.Index(i).Set(ev.Elem())
					} else {
						panic(fmt.Errorf("property value %s isn't []map[string]interface{}", opt.fullPropName))
					}
				}
				v.Set(result)
			} else {
				panic(fmt.Errorf("property value %s isn't []map[string]interface{}", opt.fullPropName))
			}
		}
	case reflect.Map:
		if t.Key().Kind() != reflect.String {
			panic(fmt.Errorf("field: %s isn't map[string]interface{}", opt.fieldName))
		}

		elemType := t.Elem()
		elemKind := elemType.Kind()

		// 首先处理使用类型转换器的场景
		if fn, ok := typeConverters[elemType]; ok {
			if mapValue, err := cast.ToStringMapStringE(propValue); err == nil {
				prefix := key + "."
				fnValue := reflect.ValueOf(fn)
				result := reflect.MakeMap(t)
				for k0, v0 := range mapValue {
					res := fnValue.Call([]reflect.Value{reflect.ValueOf(v0)})
					k0 = strings.TrimPrefix(k0, prefix)
					result.SetMapIndex(reflect.ValueOf(k0), res[0])
				}
				v.Set(result)
				return
			} else {
				panic(fmt.Errorf("property value %s isn't map[string]string", opt.fullPropName))
			}
		}

		switch elemKind {
		case reflect.Uint64, reflect.Uint32, reflect.Uint16, reflect.Uint8, reflect.Uint:
			panic(errors.New("暂未支持"))
		case reflect.Int64, reflect.Int32, reflect.Int16, reflect.Int8, reflect.Int:
			panic(errors.New("暂未支持"))
		case reflect.Float64, reflect.Float32:
			panic(errors.New("暂未支持"))
		case reflect.Bool:
			panic(errors.New("暂未支持"))
		case reflect.String:
			if mapValue, err := cast.ToStringMapStringE(propValue); err == nil {
				prefix := key + "."
				result := make(map[string]string)
				for k0, v0 := range mapValue {
					k0 = strings.TrimPrefix(k0, prefix)
					result[k0] = v0
				}
				v.Set(reflect.ValueOf(result))
			} else {
				panic(fmt.Errorf("property value %s isn't map[string]string", opt.fullPropName))
			}
		default:
			// 处理结构体字段的场景
			if mapValue, err := cast.ToStringMapE(propValue); err == nil {
				temp := make(map[string]map[string]interface{})
				trimKey := key + "."
				var ok bool

				// 将一维 map 变成二维 map
				for k0, v0 := range mapValue {
					k0 = strings.TrimPrefix(k0, trimKey)
					sk := strings.Split(k0, ".")
					var item map[string]interface{}
					if item, ok = temp[sk[0]]; !ok {
						item = make(map[string]interface{})
						temp[sk[0]] = item
					}
					item[sk[1]] = v0
				}

				result := reflect.MakeMapWithSize(t, len(temp))
				for k1, v1 := range temp {
					ev := reflect.New(elemType)
					subFullPropName := fmt.Sprintf("%s.%s", key, k1)
					bindStruct(&defaultProperties{v1}, ev.Elem(), bindOption{
						fullPropName: subFullPropName,
						fieldName:    opt.fieldName,
						allAccess:    opt.allAccess,
					})
					result.SetMapIndex(reflect.ValueOf(k1), ev.Elem())
				}

				v.Set(result)
			} else {
				panic(fmt.Errorf("property value %s isn't map[string]map[string]interface{}", opt.fullPropName))
			}
		}
	default:
		panic(errors.New(opt.fieldName + " unsupported type " + v.Kind().String()))
	}
}

// BindProperty 根据类型获取属性值，属性名称统一转成小写。
func (p *defaultProperties) BindProperty(key string, i interface{}) {
	p.BindPropertyIf(key, i, false)
}

// BindPropertyIf 根据类型获取属性值，属性名称统一转成小写。
func (p *defaultProperties) BindPropertyIf(key string, i interface{}, allAccess bool) {

	v := reflect.ValueOf(i)
	if v.Kind() != reflect.Ptr {
		panic(errors.New("参数 v 必须是一个指针"))
	}

	t := v.Type().Elem()
	s := t.Name() // 当绑定对象是 map 或者 slice 时，取元素的类型名
	if s == "" && (t.Kind() == reflect.Map || t.Kind() == reflect.Slice) {
		s = t.Elem().Name()
	}

	bindValue(p, v.Elem(), key, nil, bindOption{
		fieldName:    s,
		fullPropName: key,
		allAccess:    allAccess,
	})
}
